@@ -1,20 +1,36 @@
 """Command handlers for python processes."""
 import abc
+import importlib
 import logging
-from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type
+import os
+from base64 import b64encode
+from io import BytesIO
+from pathlib import Path
+from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union
+from zipfile import ZIP_STORED, ZipFile
 
 from django.apps import apps
 from django.conf import settings
 from django.contrib.auth import get_user_model
-from django.contrib.postgres.fields.jsonb import JSONField
-from django.db.models import ManyToManyField, Model, QuerySet
+from django.contrib.postgres.fields.jsonb import JSONField as JSONFieldb
+from django.db.models import ForeignKey, JSONField, ManyToManyField, Model, QuerySet
 
+from resolwe.flow.executors import constants
 from resolwe.flow.executors.socket_utils import Message, Response
-from resolwe.flow.models import Collection, Data, Process, Storage
+from resolwe.flow.models import (
+    Collection,
+    Data,
+    DescriptorSchema,
+    Entity,
+    Process,
+    Storage,
+)
 from resolwe.flow.models.utils import serialize_collection_relations
 from resolwe.flow.utils import dict_dot
-from resolwe.permissions.shortcuts import get_objects_for_user
-from resolwe.permissions.utils import get_full_perm
+from resolwe.permissions.models import Permission
+from resolwe.storage.connectors import connectors
+from resolwe.storage.models import FileStorage
+from resolwe.test.utils import is_testing
 from resolwe.utils import BraceMessage as __
 
 from .plugin import ListenerPlugin
@@ -27,10 +43,10 @@ UserClass = get_user_model()
 
 
 class PermissionManager:
-    """Permission manager."""
+    """Permission manager class."""
 
     def __init__(self):
-        """Initialization."""
+        """Initialize."""
         self._plugins: Dict[str, "ExposeObjectPlugin"] = dict()
 
     def add_plugin(self, plugin_class: Type["ExposeObjectPlugin"]):
@@ -99,6 +115,31 @@ class ExposeObjectPlugin(metaclass=abc.ABCMeta):
 
     full_model_name = "app_label.model_name"
 
+    def _has_permission(
+        self, user: UserClass, model: Type[Model], model_pk: int, permission_name: str
+    ):
+        """Check if contributor has requested permissions.
+
+        :raises RuntimeError: with detailed explanation when check fails.
+        """
+        object_ = model.objects.filter(pk=model_pk)
+        if not object_:
+            raise RuntimeError(
+                f"Object {model._meta.model_name} with id {model_pk} not found."
+            )
+
+        if not object_.filter_for_user(
+            user, Permission.from_name(permission_name)
+        ).exists():
+            if object_:
+                raise RuntimeError(
+                    f"No edit permission for {model._meta.model_name} with id {model_pk}."
+                )
+            else:
+                raise RuntimeError(
+                    f"Object {model._meta.model_name} with id {model_pk} not found."
+                )
+
     def can_create(self, user: UserClass, attributes: dict, data: Data):
         """Can user create the model with given attributes.
 
@@ -131,8 +172,7 @@ class ExposeObjectPlugin(metaclass=abc.ABCMeta):
         self, user: UserClass, queryset: QuerySet, data: Data
     ) -> QuerySet:
         """Filter the objects for the given user."""
-        permission_name = get_full_perm("view", queryset.model)
-        return get_objects_for_user(user, permission_name, queryset)
+        return queryset.filter_for_user(user)
 
     def can_read(self, user: UserClass, data: Data):
         """Can read model structural info.
@@ -153,30 +193,6 @@ class ExposeData(ExposeObjectPlugin):
 
     full_model_name = "flow.Data"
 
-    def _has_permission(
-        self, user: UserClass, model: Type[Model], model_pk: int, permission_name: str
-    ):
-        """Check if contributor has requested permissions.
-
-        :raises RuntimeError: with detailed explanation when check fails.
-        """
-        full_permission_name = get_full_perm(permission_name, model)
-        object_ = model.objects.filter(pk=model_pk)
-        filtered_object = get_objects_for_user(
-            user,
-            [full_permission_name],
-            object_,
-        )
-        if not filtered_object:
-            if object_:
-                raise RuntimeError(
-                    f"No permissions: {model._meta.model_name} with id {model_pk}."
-                )
-            else:
-                raise RuntimeError(
-                    f"Object {model._meta.model_name} with id {model_pk} not found."
-                )
-
     def can_create(self, user: UserClass, model_data: Dict, data: Data):
         """Can user update the given model instance.
 
@@ -188,7 +204,9 @@ class ExposeData(ExposeObjectPlugin):
             "output",
             "input",
             "tags",
+            "entity",
             "entity_id",
+            "collection",
             "collection_id",
             "name",
         }
@@ -219,8 +237,12 @@ class ExposeData(ExposeObjectPlugin):
             "output",
             "tags",
             "entity_id",
+            "entity",
             "collection_id",
+            "collection",
             "name",
+            "descriptor_schema_id",
+            "descriptor_schema",
             "descriptor",
         }
         not_allowed_keys = set(model_data.keys()) - allowed_fields
@@ -232,23 +254,29 @@ class ExposeData(ExposeObjectPlugin):
 
         if "entity_id" in model_data:
             # Check entity permissions.
-            self._has_permission(user, Process, model_data["entity_id"], "edit")
+            self._has_permission(user, Entity, model_data["entity_id"], "edit")
 
         if "collection_id" in model_data:
             # Check collection permissions.
             self._has_permission(user, Collection, model_data["collection_id"], "edit")
+        if "descriptor_schema_id" in model_data:
+            # Check DescriptorSchema permissions.
+            self._has_permission(
+                user, DescriptorSchema, model_data["descriptor_schema_id"], "view"
+            )
 
     def filter_objects(
         self, user: UserClass, queryset: QuerySet, data: Data
     ) -> QuerySet:
         """Filter the objects for the given user."""
         inputs = queryset.filter(id__in=data.parents.all())
-        permission_name = get_full_perm("view", queryset.model)
-        return (
-            get_objects_for_user(user, permission_name, queryset)
-            .distinct()
-            .union(inputs)
-        )
+        return queryset.filter_for_user(user).distinct().union(inputs)
+
+
+class ExposeDescriptorSchema(ExposeObjectPlugin):
+    """Expose the DescriptorSchema model."""
+
+    full_model_name = "flow.DescriptorSchema"
 
 
 class ExposeUser(ExposeObjectPlugin):
@@ -257,7 +285,7 @@ class ExposeUser(ExposeObjectPlugin):
     Used to read contributor data.
     """
 
-    full_model_name = "auth.User"
+    full_model_name = UserClass._meta.label
 
     def filter_objects(
         self, user: UserClass, queryset: QuerySet, data: Data
@@ -271,6 +299,34 @@ class ExposeEntity(ExposeObjectPlugin):
 
     full_model_name = "flow.Entity"
 
+    def can_create(self, user: UserClass, attributes: dict, data: Data):
+        """Can user update the given model instance.
+
+        :raises RuntimeError: if user does not have permissions to create the
+            given model.
+        """
+
+    def can_update(
+        self, user: UserClass, model_instance: Data, model_data: Dict, data: Data
+    ):
+        """Can user update the given model instance.
+
+        :raises RuntimeError: when user does not have permissions to update
+            the given model.
+        """
+        allowed_fields = {
+            "description",
+            "tags",
+            "name",
+            "descriptor",
+        }
+        not_allowed_keys = set(model_data.keys()) - allowed_fields
+        if not_allowed_keys:
+            raise RuntimeError(f"Not allowed to set {','.join(not_allowed_keys)}.")
+
+        # Check permission to modify the Entity object.
+        self._has_permission(user, Entity, model_instance.id, "edit")
+
 
 class ExposeCollection(ExposeObjectPlugin):
     """Expose the Collection model."""
@@ -294,13 +350,8 @@ class ExposeProcess(ExposeObjectPlugin):
         self, user: UserClass, queryset: QuerySet, data: Data
     ) -> QuerySet:
         """Filter the objects for the given user."""
-        processes_of_inputs = queryset.filter(data__in=data.parents.all())
-        permission_name = get_full_perm("view", queryset.model)
-        return (
-            get_objects_for_user(user, permission_name, queryset)
-            .distinct()
-            .union(processes_of_inputs)
-        )
+        processes_of_inputs = queryset.filter(data__in=data.parents.all()).distinct()
+        return queryset.filter_for_user(user).distinct().union(processes_of_inputs)
 
 
 class ExposeStorage(ExposeObjectPlugin):
@@ -353,19 +404,9 @@ class ExposeStorage(ExposeObjectPlugin):
     def filter_objects(
         self, user: UserClass, queryset: QuerySet, data: Data
     ) -> QuerySet:
-        """Filter the objects for the given user.
-
-        Snorage objects are special: we have to check if user has permission on
-        the data model. The permission name must be given in the form
-        'app_label.permission_name' otherwise wrong content type is infered
-        it the get_objects_for_user method.
-        """
-        permission_name = "flow." + get_full_perm("view", Data)
-        perms_filter = "data__pk__in"
+        """Filter the objects for the given user."""
         return (
-            get_objects_for_user(
-                user, permission_name, queryset, perms_filter=perms_filter
-            )
+            queryset.filter_for_user(user)
             .distinct()
             .union(queryset.filter(contributor=user))
         )
@@ -383,8 +424,30 @@ class PythonProcess(ListenerPlugin):
     def __init__(self):
         """Initialize plugin."""
         self._permission_manager = permission_manager
+        self._hydrate_cache: Dict[int, str] = dict()
         super().__init__()
 
+    def handle_resolve_data_path(
+        self, message: Message[int], manager: "Processor"
+    ) -> Response[str]:
+        """Return the base path that stores given data."""
+        data_pk = message.message_data
+        if data_pk not in self._hydrate_cache or is_testing():
+            mount_point = os.fspath(constants.INPUTS_VOLUME)
+            data_connectors = FileStorage.objects.get(data__pk=data_pk).connectors
+            for connector in connectors.for_storage("data"):
+                if connector in data_connectors and connector.mountable:
+                    mount_point = f"/data_{connector.name}"
+                    break
+            self._hydrate_cache[data_pk] = mount_point
+        return message.respond_ok(self._hydrate_cache[data_pk])
+
+    def handle_get_python_program(
+        self, message: Message[Tuple[str, Dict[str, Any]]], manager: "Processor"
+    ) -> Response[str]:
+        """Return the process source code for the given data object."""
+        return message.respond_ok(manager.data.process.run.get("program", ""))
+
     def handle_create_object(
         self, message: Message[Tuple[str, Dict[str, Any]]], manager: "Processor"
     ) -> Response[int]:
@@ -403,13 +466,34 @@ class PythonProcess(ListenerPlugin):
         return message.respond_ok(model.objects.create(**model_data).id)
 
     def handle_filter_objects(
-        self, message: Message[Tuple[str, Dict[str, Any]]], manager: "Processor"
+        self,
+        message: Message[
+            Union[
+                Tuple[str, str, Dict[str, Any], List[str], List[str]],
+                Tuple[str, str, Dict[str, Any], List[str]],
+            ],
+        ],
+        manager: "Processor",
     ) -> Response[List[int]]:
         """Get a list of objects based on criteria.
 
-        List is further filtered based on user permissions.
+        List is further filtered based on user permissions and sorted according
+        to the list of strings the user provides.
         """
-        app_name, model_name, filters, attributes = message.message_data
+        # Sorting was added later. For compatibility reasons handle both
+        # message types, remove the one without sorting ASAP.
+        sorting: List[str] = []
+        if len(message.message_data) == 4:
+            app_name, model_name, filters, attributes = message.message_data
+        elif len(message.message_data) == 5:
+            app_name, model_name, filters, sorting, attributes = message.message_data
+        else:
+            raise RuntimeError(
+                "Message in handle_filter_objects for the object "
+                f"{manager.data.pk} not in the correct format. Got tuple of"
+                f"length {len(message.message_data)}, expected length 4 or 5."
+            )
+
         full_model_name = f"{app_name}.{model_name}"
         model = apps.get_model(app_name, model_name)
         filtered_objects = self._permission_manager.filter_objects(
@@ -418,7 +502,9 @@ class PythonProcess(ListenerPlugin):
             model.objects.filter(**filters),
             manager.data,
         )
-        return message.respond_ok(list(filtered_objects.values_list(*attributes)))
+        return message.respond_ok(
+            list(filtered_objects.order_by(*sorting).values_list(*attributes))
+        )
 
     def handle_update_model_fields(
         self, message: Message[Tuple[str, int, Dict[str, Any]]], manager: "Processor"
@@ -455,14 +541,15 @@ class PythonProcess(ListenerPlugin):
             # only updated, other JSON fields should probably be replaced.
             # Compromise: when update is a dict, then only values in dict are
             # updates, else replaced.
-            if isinstance(model._meta.get_field(field_name), JSONField) and isinstance(
+            field_meta = model._meta.get_field(field_name)
+            if isinstance(field_meta, (JSONField, JSONFieldb)) and isinstance(
                 field_value, dict
             ):
                 update_fields.append(field_name)
                 current_value = getattr(model_instance, field_name)
                 for key, value in field_value.items():
                     dict_dot(current_value, key, value)
-            elif isinstance(model._meta.get_field(field_name), ManyToManyField):
+            elif isinstance(field_meta, ManyToManyField):
                 assert isinstance(
                     field_value, list
                 ), "Only lists may be assigned to many-to-many relations"
@@ -475,6 +562,11 @@ class PythonProcess(ListenerPlugin):
                     field.remove(*objects_to_remove)
                 if objects_to_add:
                     field.add(*objects_to_add)
+            # Set ID directly when setting foreign key relations.
+            elif isinstance(field_meta, ForeignKey):
+                field_name = f"{field_name}_id"
+                update_fields.append(field_name)
+                setattr(model_instance, field_name, field_value)
             else:
                 update_fields.append(field_name)
                 setattr(model_instance, field_name, field_value)
@@ -544,11 +636,11 @@ class PythonProcess(ListenerPlugin):
         self, message: Message[int], manager: "Processor"
     ) -> Response[List[dict]]:
         """Get relations for the given collection object."""
-        collection = get_objects_for_user(
-            manager.contributor,
-            get_full_perm("view", Collection),
-            Collection.objects.filter(id=message.message_data),
-        ).get()
+        collection = (
+            Collection.objects.filter(id=message.message_data)
+            .filter_for_user(manager.contributor)
+            .get()
+        )
         return message.respond_ok(serialize_collection_relations(collection))
 
     def handle_get_process_requirements(
@@ -556,35 +648,61 @@ class PythonProcess(ListenerPlugin):
     ) -> Response[dict]:
         """Return the requirements for the process with the given id."""
         process_id = message.message_data
-        filtered_process = get_objects_for_user(
-            manager.contributor,
-            get_full_perm("view", Process),
-            Process.objects.filter(pk=process_id),
-        )
-        process_requirements, process_slug = filtered_process.values_list(
-            "requirements", "slug"
-        ).get()
-        resources = process_requirements.get("resources", {})
+        filtered_process = Process.objects.filter(pk=process_id).filter_for_user(
+            manager.contributor
+        )[0]
+        process_limits = filtered_process.get_resource_limits()
+        process_requirements = filtered_process.requirements
+        process_requirements["resources"] = process_limits
+        return message.respond_ok(process_requirements)
 
-        # Get limit defaults and overrides.
-        limit_defaults = getattr(settings, "FLOW_PROCESS_RESOURCE_DEFAULTS", {})
-        limit_overrides = getattr(settings, "FLOW_PROCESS_RESOURCE_OVERRIDES", {})
-
-        limits = {}
-        limits["cores"] = int(resources.get("cores", 1))
-        max_cores = getattr(settings, "FLOW_PROCESS_MAX_CORES", None)
-        if max_cores:
-            limits["cores"] = min(limits["cores"], max_cores)
-
-        memory = limit_overrides.get("memory", {}).get(process_slug, None)
-        if memory is None:
-            memory = int(
-                resources.get(
-                    "memory",
-                    # If no memory resource is configured, check settings.
-                    limit_defaults.get("memory", 4096),
-                )
-            )
-        limits["memory"] = memory
+    def handle_get_self_requirements(
+        self, message: Message[int], manager: "Processor"
+    ) -> Response[dict]:
+        """Return the requirements for the process being executed."""
+        limits = manager.data.get_resource_limits()
+        process_requirements = manager.data.process.requirements
         process_requirements["resources"] = limits
         return message.respond_ok(process_requirements)
+
+    def handle_get_python_runtime(
+        self, message: Message[str], manager: "Processor"
+    ) -> Response[str]:
+        """Return the Python Process runtime.
+
+        It is gzipped and returned as b64encoded string.
+        """
+
+        zipped = BytesIO()
+        with ZipFile(file=zipped, mode="w", compression=ZIP_STORED) as zip_handle:
+            for runtime_class_name in settings.FLOW_PROCESSES_RUNTIMES:
+                module_name, class_name = runtime_class_name.rsplit(".", 1)
+                source_path = importlib.util.find_spec(module_name).origin
+                assert (
+                    source_path is not None
+                ), f"Unable to determine the source path of the module {module_name}."
+                source_dir = Path(source_path).parent
+
+                base_destination = Path(*module_name.split(".")[:-1])
+                for source_entry in source_dir.rglob("*.py"):
+                    relative_path = source_entry.relative_to(source_dir)
+                    destination = base_destination / relative_path
+                    zip_handle.write(source_entry, destination)
+
+                # Create missing __init__.py files with empty content.
+                for path in list(Path(*module_name.split(".")).parents)[1:-1]:
+                    zip_destination = path / "__init__.py"
+                    zip_handle.writestr(os.fspath(zip_destination), "")
+
+        zipped.seek(0)
+        return message.respond_ok(b64encode(zipped.read()).decode())
+
+    def handle_get_user_model_label(
+        self, message: Message[str], manager: "Processor"
+    ) -> Response[str]:
+        """Return the label of the custom user model.
+
+        It is used to connect the model in Python process runtime with the
+        custom user model.
+        """
+        return message.respond_ok(UserClass._meta.label)
