@@ -23,8 +23,8 @@ from django.db.models import Sum
 from django.db.models.functions import Coalesce
 
 from resolwe.flow.executors import constants
-from resolwe.flow.executors.prepare import BaseFlowExecutorPreparer
 from resolwe.flow.models import Data, DataDependency, Process
+from resolwe.flow.utils import get_apps_tools
 from resolwe.storage import settings as storage_settings
 from resolwe.storage.connectors import connectors
 from resolwe.storage.connectors.baseconnector import BaseStorageConnector
@@ -43,6 +43,56 @@ KUBERNETES_TIMEOUT = 30
 logger = logging.getLogger(__name__)
 
 
+def get_mountable_connectors() -> Iterable[Tuple[str, BaseStorageConnector]]:
+    """Iterate through all the storages and find mountable connectors.
+
+    :returns: list of tuples (storage_name, connector).
+    """
+    return [
+        (storage_name, connector)
+        for storage_name in storage_settings.FLOW_STORAGE
+        for connector in connectors.for_storage(storage_name)
+        if connector.mountable
+    ]
+
+
+def get_upload_dir() -> str:
+    """Get the upload path.
+
+    : returns: the path of the first mountable connector for storage
+        'upload'.
+
+    :raises RuntimeError: if no applicable connector is found.
+    """
+    for connector in connectors.for_storage("upload"):
+        if connector.mountable:
+            return f"/upload_{connector.name}"
+    raise RuntimeError("No mountable upload connector is defined.")
+
+
+def unique_volume_name(base_name: str, data_id: int) -> str:
+    """Get unique persistent volume claim name."""
+    return f"{base_name}-{data_id}"
+
+
+def sanitize_kubernetes_label(label: str, trim_end: bool = True) -> str:
+    """Make sure kubernetes label complies with the rules.
+
+    See the URL bellow for details.
+
+    https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/
+    """
+    max_length = 63
+    sanitized_label = re.sub("[^0-9a-zA-Z\-]+", "-", label).strip("-_.")
+    if len(sanitized_label) > max_length:
+        logger.warning(__("Label '%s' is too long and was truncated.", label))
+        if trim_end:
+            sanitized_label = sanitized_label[-max_length:].strip("-_.")
+        else:
+            sanitized_label = sanitized_label[:max_length].strip("-_.")
+    return sanitized_label
+
+
 class Connector(BaseConnector):
     """Kubernetes-based connector for job execution."""
 
@@ -59,19 +109,6 @@ class Connector(BaseConnector):
             "namespace", "default"
         )
 
-    def _get_upload_dir(self) -> str:
-        """Get upload path.
-
-        : returns: the path of the first mountable connector for storage
-            'upload'.
-
-        :raises RuntimeError: if no applicable connector is found.
-        """
-        for connector in connectors.for_storage("upload"):
-            if connector.mountable:
-                return f"/upload_{connector.name}"
-        raise RuntimeError("No mountable upload connector is defined.")
-
     def _prepare_environment(
         self, data: Data, listener_connection: Tuple[str, str, str]
     ) -> list:
@@ -97,24 +134,12 @@ class Connector(BaseConnector):
             ),
         }
         with suppress(RuntimeError):
-            environment["UPLOAD_DIR"] = self._get_upload_dir()
+            environment["UPLOAD_DIR"] = get_upload_dir()
 
         return [
             {"name": name, "value": str(value)} for name, value in environment.items()
         ]
 
-    def _get_upload_connector_name(self, data: Data) -> str:
-        """Get the connector for data upload.
-
-        Read the connector name from the StorageLocation class created by the
-        dispatcher.
-        """
-        return data.location.default_storage_location.connector_name
-
-    def _ebs_claim_name(self, base_name: str, data_id: int) -> str:
-        """Get unique EBS claim name."""
-        return f"{base_name}-{data_id}"
-
     def _create_configmap_if_needed(self, name: str, content: Dict, core_api: Any):
         """Create configmap if necessary."""
         try:
@@ -139,33 +164,13 @@ class Connector(BaseConnector):
 
     def _get_tools_configmaps(self, core_api):
         """Get and return configmaps for tools."""
-
-        def dict_from_directory(directory: Path) -> Dict[str, str]:
-            """Get dictionary from given directory.
-
-            File names are keys and corresponding file contents are values.
-            """
-            return {
-                entry.name: entry.read_text()
-                for entry in directory.glob("*")
-                if entry.is_file()
-            }
-
-        configmap_names = []
-        preparer = BaseFlowExecutorPreparer()
-        tools_paths = preparer.get_tools_paths()
-        for tool_path in tools_paths:
-            tool_path = Path(tool_path)
-            data = dict_from_directory(tool_path)
-            data_md5 = hashlib.md5(
-                json.dumps(data, sort_keys=True).encode()
-            ).hexdigest()
-            configmap_name = self._sanitize_kubernetes_label(
-                f"tools-{tool_path.name}-{data_md5}"
-            )
-            self._create_configmap_if_needed(configmap_name, data, core_api)
-            configmap_names.append(configmap_name)
-        return configmap_names
+        description_configmap_name = (
+            getattr(settings, "KUBERNETES_TOOLS_CONFIGMAPS", None) or "tools-configmaps"
+        )
+        configmap = core_api.read_namespaced_config_map(
+            name=description_configmap_name, namespace=self.kubernetes_namespace
+        )
+        return configmap.data.values()
 
     def _get_files_configmap_name(self, location_subpath: Path, core_api: Any):
         """Get or create configmap for files.
@@ -207,27 +212,15 @@ class Connector(BaseConnector):
         self._create_configmap_if_needed(configmap_name, data, core_api)
         return configmap_name
 
-    def _get_mountable_connectors(self) -> Iterable[Tuple[str, BaseStorageConnector]]:
-        """Iterate through all the storages and find mountable connectors.
-
-        :returns: list of tuples (storage_name, connector).
-        """
-        return (
-            (storage_name, connector)
-            for storage_name in storage_settings.FLOW_STORAGE
-            for connector in connectors.for_storage(storage_name)
-            if connector.mountable
-        )
-
     def _volumes(self, data_id: int, location_subpath: Path, core_api: Any) -> list:
         """Prepare all volumes."""
 
         def volume_from_config(volume_name: str, volume_config: dict):
             """Get configuration for kubernetes for given volume."""
             claim_name = volume_config["config"]["name"]
-            if volume_config["type"] == "ebs":
-                claim_name = self._ebs_claim_name(claim_name, data_id)
-            if volume_config["type"] in ["ebs", "persistent_volume"]:
+            if self._should_create_pvc(volume_config):
+                claim_name = unique_volume_name(claim_name, data_id)
+            if volume_config["type"] == "persistent_volume":
                 return {
                     "name": volume_name,
                     "persistentVolumeClaim": {"claimName": claim_name},
@@ -268,13 +261,20 @@ class Connector(BaseConnector):
             for volume_name, volume_config in storage_settings.FLOW_VOLUMES.items()
             if volume_config["config"]["name"] != "tools"
         ]
-        volumes += [
-            {
+
+        for storage_name, connector in get_mountable_connectors():
+            claim_name = connector.config.get("persistent_volume_claim", None)
+            volume_data = {
                 "name": connector.name,
-                "hostPath": {"path": os.fspath(connector.config["path"])},
             }
-            for storage_name, connector in self._get_mountable_connectors()
-        ]
+            if claim_name:
+                volume_data["persistentVolumeClaim"] = {"claimName": claim_name}
+            else:
+                volume_data["hostPath"] = (
+                    {"path": os.fspath(connector.config["path"])},
+                )
+            volumes.append(volume_data)
+
         return volumes
 
     def _init_container_mountpoints(self):
@@ -309,7 +309,7 @@ class Connector(BaseConnector):
                 "mountPath": f"/{storage_name}_{connector.name}",
                 "readOnly": False,
             }
-            for storage_name, connector in self._get_mountable_connectors()
+            for storage_name, connector in get_mountable_connectors()
         ]
         return mount_points
 
@@ -336,7 +336,7 @@ class Connector(BaseConnector):
                 "mountPath": f"/{storage_name}_{connector.name}",
                 "readOnly": False,
             }
-            for storage_name, connector in self._get_mountable_connectors()
+            for storage_name, connector in get_mountable_connectors()
         ]
         return mount_points
 
@@ -371,7 +371,7 @@ class Connector(BaseConnector):
                 "mountPath": f"/{storage_name}_{connector.name}",
                 "readOnly": storage_name != "upload",
             }
-            for storage_name, connector in self._get_mountable_connectors()
+            for storage_name, connector in get_mountable_connectors()
         ]
 
         mount_points += [
@@ -420,7 +420,7 @@ class Connector(BaseConnector):
                 "readOnly": True,
             },
         ]
-        for tool_index in range(len(BaseFlowExecutorPreparer().get_tools_paths())):
+        for tool_index in range(len(get_apps_tools())):
             mount_points.append(
                 {
                     "name": f"tools-{tool_index}",
@@ -429,11 +429,10 @@ class Connector(BaseConnector):
             )
         return mount_points
 
-    def _persistent_ebs_claim(
+    def _persistent_volume_claim(
         self, claim_name: str, size: int, volume_config: Dict
     ) -> Dict[str, Any]:
-        # self, persistent_claim_name: str, volume_size_in_bytes: int
-        """Prepare claim for EBS Amazon storage."""
+        """Prepare claim for persistent volume."""
         return {
             "apiVersion": "v1",
             "kind": "PersistentVolumeClaim",
@@ -449,7 +448,7 @@ class Connector(BaseConnector):
             },
         }
 
-    def _data_inputs_size(self, data: Data, safety_buffer: int = 2 ** 30) -> int:
+    def _data_inputs_size(self, data: Data, safety_buffer: int = 2**30) -> int:
         """Get the size of data inputs.
 
         Also add 10% of the volume size + 2GB as safety buffer. When having a
@@ -469,22 +468,12 @@ class Connector(BaseConnector):
         assert isinstance(inputs_size, int)
         return int(1.1 * (inputs_size + safety_buffer))
 
-    def _sanitize_kubernetes_label(self, label: str, trim_end: bool = True) -> str:
-        """Make sure kubernetes label complies with the rules.
+    def _should_create_pvc(self, volume_config: dict) -> bool:
+        """Return True if pvc for the given volume should be created."""
+        if volume_config["type"] != "persistent_volume":
+            return False
 
-        See the URL bellow for details.
-
-        https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/
-        """
-        max_length = 63
-        sanitized_label = re.sub("[^0-9a-zA-Z\-]+", "-", label).strip("-_.")
-        if len(sanitized_label) > max_length:
-            logger.warning(__("Label '%s' is too long and was truncated.", label))
-            if trim_end:
-                sanitized_label = sanitized_label[-max_length:].strip("-_.")
-            else:
-                sanitized_label = sanitized_label[:max_length].strip("-_.")
-        return sanitized_label
+        return volume_config["config"].get("create_pvc", False)
 
     def _image_mapper(self, image_name: str, mapper: Dict[str, str]) -> str:
         """Transform the image name if necessary.
@@ -508,7 +497,11 @@ class Connector(BaseConnector):
 
         # Create kubernetes API every time otherwise it will time out
         # eventually and raise API exception.
-        kubernetes.config.load_kube_config()
+        try:
+            kubernetes.config.load_kube_config()
+        except kubernetes.config.config_exception.ConfigException:
+            kubernetes.config.load_incluster_config()
+
         batch_api = kubernetes.client.BatchV1Api()
         core_api = kubernetes.client.CoreV1Api()
 
@@ -522,7 +515,7 @@ class Connector(BaseConnector):
 
         # Set resource limits.
         requests = dict()
-        limits = data.process.get_resource_limits()
+        limits = data.get_resource_limits()
 
         requests["cpu"] = limits.pop("cores")
         limits["cpu"] = requests["cpu"] + 1
@@ -539,8 +532,8 @@ class Connector(BaseConnector):
 
         requests["memory"] = 0.9 * limits["memory"]
         limits["memory"] = 1.1 * limits["memory"] + KUBERNETES_MEMORY_HARD_LIMIT_BUFFER
-        limits["memory"] *= 2 ** 20  # 2 ** 20 = mebibyte
-        requests["memory"] *= 2 ** 20
+        limits["memory"] *= 2**20  # 2 ** 20 = mebibyte
+        requests["memory"] *= 2**20
 
         resources = data.process.requirements.get("resources", {})
         network = "bridge"
@@ -603,37 +596,57 @@ class Connector(BaseConnector):
             processing_container_image, mapper
         )
 
+        affinity = {}
+        kubernetes_affinity = getattr(settings, "FLOW_KUBERNETES_AFFINITY", None)
+        if kubernetes_affinity:
+            affinity = {
+                "nodeAffinity": {
+                    "requiredDuringSchedulingIgnoredDuringExecution": {
+                        "nodeSelectorTerms": [
+                            {
+                                "matchExpressions": [
+                                    {
+                                        "key": "nodegroup",
+                                        "operator": "In",
+                                        "values": [kubernetes_affinity],
+                                    }
+                                ]
+                            }
+                        ]
+                    }
+                }
+            }
+
         job_type = dict(Process.SCHEDULING_CLASS_CHOICES)[data.process.scheduling_class]
         job_description = {
             "apiVersion": "batch/v1",
             "kind": "Job",
-            "metadata": {"name": self._sanitize_kubernetes_label(container_name)},
+            "metadata": {"name": sanitize_kubernetes_label(container_name)},
             "spec": {
                 # Keep finished pods around for ten seconds. If job is not
-                # deleted its PVC claim persists and it causes PV to stay0
+                # deleted its PVC claim persists and it causes PV to stay
                 # around.
                 # This can be changed by running a cron job that periodically
                 # checks for PVC that can be deleted.
-                "ttlSecondsAfterFinished": 10,
+                "ttlSecondsAfterFinished": 300,
                 "template": {
                     "metadata": {
-                        "name": self._sanitize_kubernetes_label(container_name),
+                        "name": sanitize_kubernetes_label(container_name),
                         "labels": {
                             "app": "resolwe",
                             "data_id": str(data.pk),
-                            "process": self._sanitize_kubernetes_label(
-                                data.process.slug
-                            ),
-                            "job_type": self._sanitize_kubernetes_label(job_type),
+                            "process": sanitize_kubernetes_label(data.process.slug),
+                            "job_type": sanitize_kubernetes_label(job_type),
                         },
                         "annotations": annotations,
                     },
                     "spec": {
+                        "affinity": affinity,
                         "hostNetwork": use_host_network,
                         "volumes": self._volumes(data.id, location_subpath, core_api),
                         "initContainers": [
                             {
-                                "name": self._sanitize_kubernetes_label(
+                                "name": sanitize_kubernetes_label(
                                     f"{container_name}-init"
                                 ),
                                 "image": communicator_image,
@@ -648,7 +661,7 @@ class Connector(BaseConnector):
                         ],
                         "containers": [
                             {
-                                "name": self._sanitize_kubernetes_label(container_name),
+                                "name": sanitize_kubernetes_label(container_name),
                                 "image": processing_container_image,
                                 "resources": {"limits": limits, "requests": requests},
                                 "securityContext": security_context,
@@ -663,7 +676,7 @@ class Connector(BaseConnector):
                                 ),
                             },
                             {
-                                "name": self._sanitize_kubernetes_label(
+                                "name": sanitize_kubernetes_label(
                                     f"{container_name}-communicator"
                                 ),
                                 "image": communicator_image,
@@ -691,14 +704,14 @@ class Connector(BaseConnector):
 
         processing_name = constants.PROCESSING_VOLUME_NAME
         input_name = constants.INPUTS_VOLUME_NAME
-        if storage_settings.FLOW_VOLUMES[processing_name]["type"] == "ebs":
-            claim_name = self._ebs_claim_name(
+        if self._should_create_pvc(storage_settings.FLOW_VOLUMES[processing_name]):
+            claim_name = unique_volume_name(
                 storage_settings.FLOW_VOLUMES[processing_name]["config"]["name"],
                 data.id,
             )
-            claim_size = limits.pop("storage", 200) * (2 ** 30)  # Default 200 gibibytes
+            claim_size = limits.pop("storage", 200) * (2**30)  # Default 200 gibibytes
             core_api.create_namespaced_persistent_volume_claim(
-                body=self._persistent_ebs_claim(
+                body=self._persistent_volume_claim(
                     claim_name,
                     claim_size,
                     storage_settings.FLOW_VOLUMES[processing_name]["config"],
@@ -707,14 +720,14 @@ class Connector(BaseConnector):
                 _request_timeout=KUBERNETES_TIMEOUT,
             )
         if input_name in storage_settings.FLOW_VOLUMES:
-            if storage_settings.FLOW_VOLUMES[input_name]["type"] == "ebs":
+            if self._should_create_pvc(storage_settings.FLOW_VOLUMES[input_name]):
                 claim_size = self._data_inputs_size(data)
-                claim_name = self._ebs_claim_name(
+                claim_name = unique_volume_name(
                     storage_settings.FLOW_VOLUMES[input_name]["config"]["name"],
                     data.id,
                 )
                 core_api.create_namespaced_persistent_volume_claim(
-                    body=self._persistent_ebs_claim(
+                    body=self._persistent_volume_claim(
                         claim_name,
                         claim_size,
                         storage_settings.FLOW_VOLUMES[input_name]["config"],
@@ -723,6 +736,7 @@ class Connector(BaseConnector):
                     _request_timeout=KUBERNETES_TIMEOUT,
                 )
 
+        logger.debug(f"Creating namespaced job: {job_description}")
         batch_api.create_namespaced_job(
             body=job_description,
             namespace=self.kubernetes_namespace,
@@ -761,21 +775,27 @@ class Connector(BaseConnector):
         )
 
     def cleanup(self, data_id: int):
-        """Remove the EBS storage used by the executor."""
-        kubernetes.config.load_kube_config()
+        """Remove the persistent volume claims created by the executor."""
+        try:
+            kubernetes.config.load_incluster_config()
+        except kubernetes.config.config_exception.ConfigException:
+            kubernetes.config.load_kube_config()
+
         core_api = kubernetes.client.CoreV1Api()
         claim_names = [
-            self._ebs_claim_name(type_, data_id)
+            unique_volume_name(type_, data_id)
             for type_ in [
                 constants.PROCESSING_VOLUME_NAME,
                 constants.INPUTS_VOLUME_NAME,
             ]
+            if type_ in storage_settings.FLOW_VOLUMES
+            and self._should_create_pvc(storage_settings.FLOW_VOLUMES[type_])
         ]
-        for ebs_claim_name in claim_names:
-            logger.debug("Kubernetes: removing claim %s.", ebs_claim_name)
+        for persistent_claim_name in claim_names:
+            logger.debug("Kubernetes: removing claim %s.", persistent_claim_name)
             with suppress(kubernetes.client.rest.ApiException):
                 core_api.delete_namespaced_persistent_volume_claim(
-                    name=ebs_claim_name,
+                    name=persistent_claim_name,
                     namespace=self.kubernetes_namespace,
                     _request_timeout=KUBERNETES_TIMEOUT,
                 )
