@@ -1,6 +1,7 @@
 """Storage connector."""
 import abc
 import copy
+from enum import Enum, auto
 from inspect import getfullargspec
 from os import PathLike
 from pathlib import PurePath
@@ -11,6 +12,15 @@ from wrapt import decorator
 DEFAULT_CONNECTOR_PRIORITY = 100
 
 
+class ConnectorType(Enum):
+    """Type of storage connector."""
+
+    LOCAL = auto()
+    S3 = auto()
+    GCS = auto()
+    UNKNOWN = auto()
+
+
 @decorator
 def validate_url(wrapped, instance, args, kwargs):
     """Enforces argument named "url" to be relative path.
@@ -54,6 +64,7 @@ class BaseStorageConnector(metaclass=abc.ABCMeta):
 
     REQUIRED_SETTINGS = ["Connector must override REQUIRED_SETTINGS"]
     CHUNK_SIZE = 8 * 1024 * 1024  # 8 MB
+    CONNECTOR_TYPE = ConnectorType.UNKNOWN
 
     def __init__(self, config: dict, name: str):
         """Connector initialization."""
@@ -156,13 +167,18 @@ class BaseStorageConnector(metaclass=abc.ABCMeta):
 
     @abc.abstractmethod
     def push(
-        self, stream: BinaryIO, url: Union[str, PathLike], chunk_size: int = CHUNK_SIZE
+        self,
+        stream: BinaryIO,
+        url: Union[str, PathLike],
+        chunk_size: int = CHUNK_SIZE,
+        hashes: Dict[str, str] = {},
     ):
         """Push data from the stream to the given URL.
 
         :param stream: given stream.
         :param url: where the data in the stream will be stored.
         :param chunk_size: the chunk_size to use.
+        :param hashes: the hashes to set (as metadata) on the uploaded object.
         """
         raise NotImplementedError
 
@@ -323,3 +339,10 @@ class BaseStorageConnector(metaclass=abc.ABCMeta):
     def __repr__(self):
         """Get string representation."""
         return self.__str__()
+
+    def temporary_credentials(self, prefix: str, duration: int = 900, **kwargs) -> Dict:
+        """Get the temporary credentials.
+
+        The default implementation returns empty dictionary.
+        """
+        return {}
